/* TCP Multipoint Control Unit (tcpmcu)
 *
 * Brent Baccala <baccala@freesoft.org>
 *
 * This program is public domain.
 * Please send any useful enhancements to the author.
 *
 * February 2001
 *
 * $Log: tcpmcu.c,v $
 * Revision 1.14  2010/04/21 04:56:03  baccala
 * updated
 *
 * Revision 1.13  2002/01/26 17:06:58  baccala
 * Make port number optional and added -b (block) option
 *
 * Revision 1.12  2002/01/26 16:22:44  baccala
 * Fixed spelling error
 *
 * Revision 1.11  2002/01/26 16:17:42  baccala
 * Receive urgent data in-line.  This fixes the problem of bytes being
 * dropped when TCP flags them urgent, but doesn't propagate the urgent
 * notifications on to the other side of the connection.  We'll see if
 * it's a problem...
 *
 * Revision 1.10  2002/01/25 00:10:30  baccala
 * Added login banner to state what you're connecting to
 *
 * Revision 1.9  2001/09/28 07:55:19  baccala
 * Added support for CORBA (ORBit implementation)
 *
 * Compile with -DUseORBIT to enable CORBA controls; omitting this directive
 * compiles the program without any need for ORBit librarys or include files
 *
 * Revision 1.8  2001/07/20 22:01:44  baccala
 * Use a linked list of structures to track client connections instead of
 * bits in an fdset
 *
 * Revision 1.7  2001/07/20 19:01:38  baccala
 * Get rid of writefds and exceptfds on select; they weren't used and could
 * cause the program to poll if they returned true
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdarg.h>
#include <netdb.h>
#include <string.h>
#include <strings.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include <arpa/inet.h>

#ifdef UseORBIT
#include "orb/orbit.h"
#include "tcpmcu.h"
#endif

struct clientconn {
  struct sockaddr_in client_addr;
  int fd;
  char *name;
  int inputEnabled;
  struct clientconn *next;
};

struct clientconn *clients = NULL;

static int debuglevel = 0;

static inline void debug(int level, const char *format, ...) {
  va_list ap;

  va_start(ap, format);
  if (level <= debuglevel) {
    vfprintf(stderr, format, ap);
  }
  va_end(ap);
}

void usage()
{
  fprintf(stderr, "Usage: tcpmcu [options] hostname [port]\n");
  fprintf(stderr, "Options:\n");
  fprintf(stderr, "   -p portnum       port to listen on (default dynamic)\n");
  fprintf(stderr, "   -d debuglev      debug level (default 0)\n");
  fprintf(stderr, "   -b               block write access on new connections\n");
  fprintf(stderr, "   -r               restrict access to localhost\n");
  exit(1);
}


fd_set CORBAfds;
int maxCORBAfd = -1;

int defaultInputState = 1;

#ifdef UseORBIT

CORBA_Environment ev;
CORBA_ORB orb;

/* This is so we can get out a valid IOR later... */
tcpmcu tcpmcu_client = CORBA_OBJECT_NIL;


static userList *
do_listusers(PortableServer_Servant servant, CORBA_Environment *ev)
{
  userList *ul = userList__alloc();

  if (clients != NULL) {
    int usercount=0;
    struct clientconn *p;
    for (p=clients; p; p=p->next) {
      usercount ++;
    }

    ul->_buffer = CORBA_sequence_user_allocbuf(usercount);
    ul->_maximum = usercount;
    ul->_length = usercount;

    usercount = 0;
    for (p=clients; p; p=p->next, usercount++) {
      ul->_buffer[usercount].name = "";
      ul->_buffer[usercount].ipaddr
	= strdup(inet_ntoa(p->client_addr.sin_addr));
      ul->_buffer[usercount].connection = p->fd;
      ul->_buffer[usercount].inputEnabled = p->inputEnabled;
    }
  }

  return ul;
}

void
do_enableinput(PortableServer_Servant servant, short connection,
	       CORBA_Environment *ev)
{
  struct clientconn *p;

  for (p=clients; p; p=p->next) {
    if (p->fd == connection) p->inputEnabled = 1;
  }
}

void
do_disableinput(PortableServer_Servant servant, short connection,
		CORBA_Environment *ev)
{
  struct clientconn *p;

  for (p=clients; p; p=p->next) {
    if (p->fd == connection) p->inputEnabled = 0;
  }
}

/*
 * I have **no** idea what this bit does
 */
PortableServer_ServantBase__epv base_epv = {
  NULL,
  NULL,
  NULL
};
POA_tcpmcu__epv tcpmcu_epv = { NULL,
			       do_listusers, do_enableinput, do_disableinput };
POA_tcpmcu__vepv poa_tcpmcu_vepv = { &base_epv, &tcpmcu_epv };
POA_tcpmcu poa_tcpmcu_servant = { NULL, &poa_tcpmcu_vepv };

static void
orb_add_connection(GIOPConnection *cnx)
{
  int fd = GIOP_CONNECTION_GET_FD(cnx);

  FD_SET(fd, &CORBAfds);

  if (fd > maxCORBAfd) maxCORBAfd = fd;
}

static void
orb_remove_connection(GIOPConnection *cnx)
{
  int fd = GIOP_CONNECTION_GET_FD(cnx);

  FD_CLR(fd, &CORBAfds);
}


static void
ORBIT_init(int *argcp, char *argv[])
{
  PortableServer_ObjectId objid = {0, sizeof("tcpmcu"), "tcpmcu"};
  PortableServer_POA poa;


  CORBA_exception_init(&ev);
  IIOPAddConnectionHandler = orb_add_connection;
  IIOPRemoveConnectionHandler = orb_remove_connection;
  orb = CORBA_ORB_init(argcp, argv, "orbit-local-orb", &ev);

  POA_tcpmcu__init(&poa_tcpmcu_servant, &ev);

  poa = (PortableServer_POA)CORBA_ORB_resolve_initial_references(orb, "RootPOA", &ev);
  PortableServer_POAManager_activate(PortableServer_POA__get_the_POAManager(poa, &ev), &ev);
  PortableServer_POA_activate_object_with_id(poa,
					     &objid, &poa_tcpmcu_servant, &ev);

  tcpmcu_client = PortableServer_POA_servant_to_reference(poa,
							  &poa_tcpmcu_servant,
							  &ev);
  if (tcpmcu_client) {
    char * retval;
    FILE * ofp;
    char filename[32];

    retval = CORBA_ORB_object_to_string(orb, tcpmcu_client, &ev);

    snprintf(filename, sizeof(filename), "/tmp/tcpmcu.ior");
    ofp = fopen(filename, "w");

    fprintf(ofp,"%s", retval);
    fclose(ofp);

    /* fprintf(stderr, "%s\n", retval); */

    CORBA_free(retval);
  }

  ORBit_custom_run_setup(orb, &ev);

}

#endif


int main(int argc, char *argv[])
{
  char option;
  int listenport = 0;
  int restrictaccess = 0;
  int passive, target = -1;
  struct sockaddr_in passive_addr, target_addr;
  struct hostent *target_hostent;
  extern int optind, opterr, optopt;
  char buffer[BUFSIZ];
  int one=1;


  /* Parse options */

  while ((option = getopt(argc, argv, "bd:p:r")) != -1) {
    switch (option) {
    case 'p':
      listenport = atoi(optarg);
      break;

    case 'd':
      debuglevel = atoi(optarg);
      break;

    case 'b':
      defaultInputState = 0;
      break;

    case 'r':
      restrictaccess = 1;
      break;

    case '?':
      usage();
    }
  }

  FD_ZERO(&CORBAfds);
#ifdef UseORBIT
  ORBIT_init(&argc, argv);
#endif

  if ((optind+2 != argc) && (optind+1 != argc)) {
    usage();
  }

  /* Use the specified hostname/port to build target_addr */

  if ((target_hostent = gethostbyname(argv[optind])) == NULL) {
    perror("gethostbyname");
  }

  bzero(&target_addr, sizeof(target_addr));
  target_addr.sin_family = AF_INET;
  if (optind+2 == argc) {
    target_addr.sin_port = htons(atoi(argv[optind+1]));
  } else {
    target_addr.sin_port = htons(23);
  }
  memcpy(&target_addr.sin_addr, *target_hostent->h_addr_list,
	 sizeof(struct in_addr));

  /* Now open a passive socket and listen on it */

  if ((passive = socket(PF_INET, SOCK_STREAM, 0)) == -1) {
    perror("socket");
    exit(1);
  }

  bzero(&passive_addr, sizeof(passive_addr));
  passive_addr.sin_family = AF_INET;
  passive_addr.sin_port = htons(listenport);
  if (restrictaccess) {
    inet_aton("127.0.0.1", &passive_addr.sin_addr);
  }

  if (bind(passive, (struct sockaddr *) &passive_addr, sizeof(passive_addr)) == -1) {
    perror("bind");
    exit(1);
  }

  if (listen(passive, 5) == -1) {
    perror("listen");
    exit(1);
  }

  if (listenport == 0) {
    int sockaddrlen = sizeof(passive_addr);
    if (getsockname(passive, (struct sockaddr *) &passive_addr, &sockaddrlen) == -1) {
      perror("getsockname");
    } else {
      listenport = ntohs(passive_addr.sin_port);
    }
  }

  fprintf(stderr, "Listening on port %d\n", listenport);

  /* Main loop - sit and wait for events on our sockets */

  while (1) {
    struct clientconn *p;
    fd_set readfds;
    int maxfd;

    memcpy(&readfds, &CORBAfds, sizeof(CORBAfds));
    maxfd = maxCORBAfd;

    FD_SET(passive, &readfds);
    if (passive > maxCORBAfd) maxfd = passive;

    for (p = clients; p; p=p->next) {
      FD_SET(p->fd, &readfds);
      if (p->fd > maxfd) maxfd=p->fd;
    }

    if (target != -1) {
      FD_SET(target, &readfds);
      if (target > maxfd) maxfd=target;
    }

    select(maxfd+1, &readfds, NULL, NULL, NULL);

#ifdef UseORBIT
    giop_main_iterate(FALSE);
#endif

    /* "passive" is readable - this is an incoming connection */

    if (FD_ISSET(passive, &readfds)) {
      int datafd;
      struct sockaddr_in incoming_addr;
      int addrlen = sizeof(incoming_addr);

      if ((datafd = accept(passive, (struct sockaddr *) &incoming_addr, &addrlen)) == -1) {
	perror("accept");
      } else {

	debug(2, "Accepting connection on fd %d from %s\n",
	      datafd, inet_ntoa(incoming_addr.sin_addr));

	p = (struct clientconn *) malloc(sizeof(struct clientconn));
	memcpy(&p->client_addr, &incoming_addr, sizeof(struct sockaddr_in));
	p->fd = datafd;
	p->inputEnabled = defaultInputState;
	p->next = clients;
	clients = p;

	/* Receive urgent data in-line. */

	if (setsockopt(datafd, SOL_SOCKET, SO_OOBINLINE, &one, sizeof(one))
	    == -1) {
	  perror("setsockopt datafd SO_OOBINLINE");
	}

	if (target != -1 || !defaultInputState) {

	  /* We're already connected to a target.  Fake TELNET options. */

	  buffer[0] = 255;	/* IAC */
	  buffer[1] = 251;	/* WILL */
	  buffer[2] = 1;	/* ECHO */

	  buffer[3] = 255;	/* IAC */
	  buffer[4] = 251;	/* WILL */
	  buffer[5] = 3;	/* SUPRESS GO AHEAD */

	  write(datafd, buffer, 6);

	}

	if (target == -1) {

	  /* We're not connected to target.  Connect now */

	  if ((target = socket(PF_INET, SOCK_STREAM, 0)) == -1) {
	    perror("socket");
	    exit(1);
	  }

	  if (connect(target, (struct sockaddr *) &target_addr,
		      sizeof(target_addr)) < 0) {
	    perror("connect");
	    exit(1);
	  }

	  /* Receive urgent data in-line. */

	  if (setsockopt(target, SOL_SOCKET, SO_OOBINLINE, &one, sizeof(one))
	      == -1) {
	    perror("setsockopt target SO_OOBINLINE");
	  }

	  debug(2, "Connected to target on fd %d\n", target);

	}

	write(datafd, buffer,
	      sprintf(buffer, "Connected to %s:%d on fd %d\r\n",
		      inet_ntoa(target_addr.sin_addr),
		      ntohs(target_addr.sin_port), datafd));

      }
    }

    /* "target" is readable - read from it and write to all data sockets */

    if ((target != -1) && FD_ISSET(target, &readfds)) {
      int len;

      debug(3, "Reading from target (%d)\n", target);

      if ((len = read(target, buffer, sizeof(buffer))) == -1) {
	perror("read");
      } else if (len == 0) {

	/* End of file - socket closed - drop all the data sockets and reset */

	debug(2, "target (%d) closed by remote\n", target);
	close(target);
	target = -1;

	for (p = clients; p; ) {
	  struct clientconn *next = p->next;

	  debug(2, "dropping datafd %d\n", p->fd);
	  close(p->fd);
	  free(p);
	  p = next;
	}
	clients = NULL;

      } else {

	for (p = clients; p; p=p->next) {
	  write(p->fd, buffer, len);
	}
      }
    }

    /* Run through the data sockets and see if any of them are readable.
     * If so, read from it and write to target.
     */

    for (p=clients; p; p=p->next) {
      if (FD_ISSET(p->fd, &readfds)) {
	int len;
	debug(3, "Reading on fd %d\n", p->fd);
	if ((len = read(p->fd, buffer, sizeof(buffer))) == -1) {
	  perror("read");
	} else if (len == 0) {

	  /* End of file - drop this datafd off our list */
	  debug(2, "fd %d closed by remote\n", p->fd);
	  close(p->fd);
	  if (clients == p) {
	    clients = p->next;
	  } else {
	    struct clientconn *n;
	    for (n=clients; n; n=n->next) {
	      if (n->next == p) {
		n->next = p->next;
	      }
	    }
	  }
	  free(p);

	  /* No data connections left?  Then drop target */
	  if (clients == NULL) {
	    debug(2, "disconnecting from target (%d)\n", target);
	    close(target);
	    target = -1;
	  }

	} else {
	  if (p->inputEnabled) write(target, buffer, len);
	}
      }
    }
  }
}
